(* :Title: Object-oriented Programming *)

(* :Name: ClassesWithPoint` *)

(* :Author: Roman E. Maeder, June 1993. *)

(* :Summary:
This packages provides the support for object-oriented programming in Mathematica.
*)
(****************************************)
(* Syntax object.method active 9/00*)
(****************************************)
(* two functions added by J.Schumacher and J.Dicker in 9/99 *)
(* newDelayed and setDelayed *)

(* one Method of Class Object added by J.Dicker in 10/99 *)
(* giveObjectData *)

(* :Context: ClassesWithPoint` *)

(* :Package Version: 1.1 *)

(* :Copyright: Copyright 1993, Roman E. Maeder.

   Permission is granted to distribute verbatim copies of this package
   together with any of your packages that use it, provided the following
   acknowledgement is printed in a standard place:
 
	"Classes.m is distributed with permission by Roman E. Maeder."
  
   The newest release of Classes.m is available through MathSource.
*)

(* :History:
   Improved efficiency, June 1993..
   Version 1.0 for the Mathematica Journal, October 1992.
*)

(* :Keywords: object-oriented programming, classes, methods, objects *)

(* :Source: 
    Maeder, Roman E.. 1993. Object-oriented Programming.
        The Mathematica Journal, 3(1).
*)

(* :Mathematica Version: 2.2 *)

(* :Limitation:
   There are no checks for the correct number of arguments of methods.
*)

(* :Discussion: 
   This version of the package is more efficient than the original one
   from the Journal. Objects use less space and are freed
   automatically, as soon as they are no longer referenced.  Therefore,
   delete[obj] should normally not be used any more.  The external
   representation of objects has been changed.  The functionality is
   the same, however.
 

*)

BeginPackage["ClassesWithPoint`"]

Class::usage = "Class[class, superclass, variables, methods] defines
	a new class as a subclass of superclass.
	Class[object] gives the class of an object."
ClassQ::usage = "ClassQ[symbol] is True, if symbol is a class."
Methods::usage = "Methods[class] gives the list of methods of class."
InstanceVariables::usage = "InstanceVariables[class] gives the
	list of instance variables of class."
SuperClass::usage = "SuperClass[class] gives the superclass of class."
new::usage = "new[class, args...] generates a new object of class. Any
	method with name 'new' is called to initialize the new object."
delete::usage = "delete[obj] deletes an object (for special occasions only)."
self::usage = "self denotes the object inside methods."
super::usage = "super denotes the object as a member of its superclass."
isa::usage = "isa[obj, class] is true if obj belongs to class
	or a subclass of it."
NIM::usage = "NIM[self, <method>]& can be used as body of a
	pure virtual method."

Object::usage = "Object is the root class."

Object::nim = "Method `1` not implemented for class `2`."


SetAttributes[newDelayed,{HoldAll,SequenceHold}];

SetAttributes[setDelayed,{HoldAll,SequenceHold}];

SetAttributes[noDelayedIndex,{HoldAll,SequenceHold}];

newDelayed::usage=
       "newDelayed[class,par] where par can be one or more parameters. It works 
        similar to new, but all given parameters are processed in `delayed` mode.
        par has to be atomic or simple list.";
        
setDelayed::usage = 
       "setDelayed[setMethod, class, par] where par can be one or more parameters. 
        It works similar to the usual `set` functions, but all given parameters are 
        processed in `delayed` mode. par has to be atomic or simple list.";

noDelayedIndex::usage=
       "noDelayedIndex[toVariable,fromVariable,index]; toVariable has to be an array,
        fromVariable a list, index an integer. What the function does is: 
        toVariable[index] := fromVariable[[>>>index<<<]]]; Where >>>index<<< is the
        evaluated index, i.e. the value of index.";

(*
giveObjectData::usage=
       "giveObjectData[object] returns a nested list of Strings that contain the data of
       all instance Variables including the data of imbedded objects.";
*)

Begin["`Private`"]

context = $Context

(* private rules for instances *) {variables}
(* private class methods *)       {methodHandler}
(* other private symbols *)       {raise}

ClassQ[_] := False (* default *)

Class[ class_Symbol,
       superclass_?ClassQ,
       variables:{_Symbol...},
       methods:{{_Symbol, _Function}...}|{}
     ] :=
    Module[{apply, standard,
            localmethods, allvariables, messages,
            methodQ},

        standard = { (* standard methods *)
            {Class,    class&},
            {isa, (class===#1 || isa[super, #1])&} };
        localmethods = Join[standard, methods];
        messages = Union[ Join[First /@ localmethods,
                               Methods[superclass]] ];
        allvariables = Join[variables, InstanceVariables[superclass]];
        (* class methods *)
        class/: Methods[class] = messages;
        class/: InstanceVariables[class] = allvariables;
        class/: SuperClass[class] = superclass;

        With[{ivars = Hold @@ allvariables,
              nvars = -Length[allvariables],
              localnames = First /@ localmethods,
              cookie = ToExpression[ context <> ToString[class] ],
              allvariables = allvariables},

          (* the head used for objects of this class *)
          SetAttributes[cookie, HoldAll];

          (* aux predicate for apply and message passing *)
          (methodQ[#] = True)& /@ messages;
          methodQ[_] = False;

          (* definitions of aux method applicator ``apply'' *)
          Apply[
            (apply[#1, obj_] :=
              With[{lvars = Take[Hold @@ obj, nvars]},
                #2 /.
                 List @@ Thread[ ivars :> lvars, Hold] /.
                 { self -> obj, super -> raise[obj, superclass] }
              ])&,
            localmethods, {1} ];
          (* inheritance, if not local method *)
          apply[f_, obj_] := methodHandler[superclass][f, obj];
          class/: methodHandler[class] = apply;

          (* message passing *)
          cookie/: (f_Symbol?methodQ)[obj_cookie, args___] :=
                   apply[f, obj][args];
          (* super *)
          raise/: (f_Symbol?methodQ)[raise[obj_, class], args___] :=
                   apply[f, obj][args];

          (* create instances of this class *)
          class/: new[class, init___] :=
            Module[{obj, syms = Unique[allvariables]},
	      SetAttributes[Evaluate[syms], Temporary];
              obj = cookie @@ syms;
              new[obj, init]; (* call any constructor defined *)
              obj
            ];
          (* formatting *)
          Format[obj_cookie] := StringForm["-`1`-", class]
        ];
        class/: ClassQ[class] = True; (* seal of approval *)
        class
    ]

(* added by Jochen Dicker in 10/99 *)
  
  valueCutIfTooBig[value_] := If[ByteCount[value]>1024,">> Data more than 1kByte <<",
                                 ToString[value],ToString[value]];

  instanceDataToString[varList_] := Module[{posList},
    If[isa[Last[varList],Object]===True, 
      giveObjectData[Last[varList]],
      (posList=StringPosition[ToString[First[varList]],"`"];
       If[(Length[posList]>0)===True,
         StringDrop[ToString[First[varList]],posList//Last //Last]<>" = "<>
           ClassesWithPoint`Private`valueCutIfTooBig[Last[varList]], 
	 ToString[First[varList]]<>" = "<>ClassesWithPoint`Private`valueCutIfTooBig[Last[varList]]
       ] (* end if Length *)
      )
    ] (* end if isa *)
  ];

(* end added JD *)

(* the class Object: we need a dummy superclass *)

Block[{noClass},
  noClass/: Methods[noClass] = {};
  noClass/: InstanceVariables[noClass] = {};
  noClass/: methodHandler[noClass] = badmessage;
  noClass/: SuperClass[noClass] = noClass;
  noClass/: ClassQ[noClass] = True;



  Class[ Object, noClass, {},
    {{new,               self&},
     {delete,            (Remove @@ self; Null)&},
     {isa,               #1===Object&},
     {Methods,           Methods[Class[self]]&},
     {InstanceVariables, InstanceVariables[Class[self]]&},
     {SuperClass,        SuperClass[Class[self]]&},
     {NIM,		 Message[Object::nim, #1, Class[self]]&}
     
     (* added by Jochen Dicker in 10/99 *)
     (*
     {giveObjectData,
        Module[{vars=InstanceVariables[self],len=Length[self],varList},
          varList=Transpose[{vars,Table[self[[i]],{i,1,len}]}];
	  {"Data of class: "<>ToString[Class[self]],
            ClassesWithPoint`Private`instanceDataToString /@ varList}
	]&
     }
     *)
    }
  ]
]

  noDelayedIndex[toVariable_,fromVariable_,index_]:=Module[{evaluatedIndex},
    evaluatedIndex=index;
    toVariable[index]:=fromVariable[[evaluatedIndex]];
  ];
		
  setDelayed[setMethod_,classPar_,par__]:=Module[{t,i,return},
    If[Head[par]===List, 
     (return=setMethod[classPar,Array[t,Length[par]]];
      Do[(noDelayedIndex[t,par,i];),{i,Length[par]}];),
    
     (setMethod[classPar,t];t:=par;)];
     return
  ];

  newDelayed[classPar_,par__]:=Module[{t,obj,i},
    If[Head[par]===List, 
      (obj=new[classPar,Array[t,Length[par]]];				
       Do[(noDelayedIndex[t,par,i];),{i,Length[par]}];),
       
      (obj=new[classPar,t];t:=par;);];
    obj
  ];
  
End[]

Protect[ Class, Methods, InstanceVariables, SuperClass,
         new, delete, isa, Object ]

EndPackage[];

(* instead of method[object,parameter] you can use now: object.method[parameter] *)
(* 27.09.00 Letay: Vereinigung von support und Jochens Idee *)
(* Matrixprodukt funktioniert auch :)*)
Unprotect[Dot];
SetAttributes[Dot, HoldRest];
Dot[object_, new[parameters___]] :=
	new[object, parameters];
(*Dot[(object_)?((ListQ[#])&), method_] :=Dot[Evaluate[Sequence[object,method]]];*)
Dot[(object_)?((ClassQ[Class[#]])&), method_[parameters___]] :=
	method[object, parameters];
Protect[Dot];	

(*Dot[object_?(Function[x,isa[Evaluate[x],Object]===True]), method_[parameters___]] :=*)
(* 27.09.00 Letay: von support korrigierte Version *)
(*
Unprotect[Dot];
SetAttributes[Dot, HoldRest];
(object_)?ObjectQ . ((method_)?ObjectQ)[
     parameters___] := method[object, parameters];
Protect[Dot];
*)
(* 30.7.98 from: Object-Oriented Programmming, p. 337 *)
(* Leider kann man so das Matrizenprodukt, was auch durch Dot angegeben ist, nicht mehr nutzen :( *)
(*
Unprotect[Dot];
SetAttributes[Dot, HoldRest];
Dot[object_, method_[parameters___]] :=
	method[object, parameters];
Protect[Dot];	
*)
