(* :Title: Fixed Length Number Formatting - A Hack *)

(* :Context: FixedNumberForm`*)

(* :Author: John M. Novak *)

(* :Summary: 
	This packed introduces the hack FixedNumberForm.  It allows
	numbers to be formatted with a fixed length mantissa and
	exponent.  Numbers are also expressed in the "e" form.
	It is intended for output to files for import into
	applications that only understand this format...
*)

(* :Package Version: 1.0 *)

(* :Mathematica Version: 2.0 *)

(* :History:
	V.1.0, May 92 by John M. Novak
*)

(* :Keywords:
	fixed length numbers, exponent format
*)

(* :Warnings:
	All numbers will be converted to the precision
	specified (or machine precision, whichever is greater.)
	This also includes symbolic numbers (e.g., Pi, E) and
	integers.
*)

(* :Discussion:
	This is more of a hack than anything.  It's principle use
	should be limited to formatting numbers for export (where
	dealing with the conversion of integers and symbolic
	numbers isn't a problem.)  It must be emphasized that this
	conversion does modify the actual expression, not just
	the formatting (to do this just in the formatting would
	require that I be a good deal more clever... :-)
	
	The basic syntax is
	FixedNumberForm[expr, mantissalength, exponentlength]
	where all numbers will be changed to mantissalength
	precision, and will be formatted as <mantissa>e<exponent>.
*)

BeginPackage["FixedNumberForm`"]

FixedNumberForm::usage =
"FixedNumberForm[expr, mantissalength, exponentlength, options] converts
numbers in expr to mantissalength precision, and formats them
in a fixed length format. I.e. 1.000e+01.
if the option mathExpression->True an string is returned which 
can be converted into an Mathematica number via ToExpression. I.e. 1.000*^01";

mathExpression::usage = "Option of FixedNumberForm. If the option mathExpression->True 
an string is returned which can be converted into an Mathematica number via ToExpression. I.e. 1.000*^01";

Options[FixedNumberForm] = {mathExpression->False};

Begin["`Private`"]

Off[StringTake::take]

padmantissa[num_, length_, opts___] :=
  Module[{str = num, zeros = length - (StringLength[num] - 1), mathMode, pSign},
    mathMode = mathExpression /. {opts} /. Options[FixedNumberForm];
    If[mathMode , pSign = " " , pSign = "+"];
      If[StringTake[num,1] === "-", zeros = zeros + 1,
	str = pSign<>str];
	str <> Table["0",{zeros}]
      ]

padexponent[exp_, length_, opts___] :=
  Module[{str = exp, sign, zeros, psign, mathMode},
    mathMode = mathExpression /. {opts} /. Options[FixedNumberForm];
    If[mathMode,pSign="",pSign="+"];
      If[StringTake[str,1] === "-",
        sign = "-"; str = StringDrop[str,1],
	sign = pSign
      ];
      zeros = length - StringLength[str];
      If[zeros < 0,
	Message[FixedNumberForm::truncexp];
	str = StringTake[str, length];
	zeros = 0
      ];
      sign <> Table["0",{zeros}] <> str
   ]

FixedNumberForm::truncexp =
"Warning: length of exponent was too great for field,
truncating...";

FixedNumberForm[expr_, 0, 0, opts___] := 
  NumberForm[N[expr, 0], {Infinity, 0}, NumberFormat -> (ToString[#1] &), 
    NumberPoint -> "", ExponentFunction -> (Null &)]

FixedNumberForm[expr_, digits_Integer:6, 0, opts___] :=
  NumberForm[N[expr, digits], {Infinity, digits}, 
    NumberFormat -> (StringJoin[#1, 
      Table["0", {digits - StringLength[#1] + StringPosition[#1, "."][[1, 1]]}]] &), 
    ExponentFunction -> (Null &)]

FixedNumberForm[expr_, digits_Integer:6, expdigits_Integer:2, opts___] := 
  Module[{expSign, mathMode},
    mathMode = mathExpression /. {opts} /. Options[FixedNumberForm];
    If[mathMode,expSign="*^",expSign="e"];
    ScientificForm[N[expr,digits], digits,
      NumberFormat -> ((SequenceForm[padmantissa[#1,digits,opts], expSign, padexponent[#3,expdigits, opts]])&)]
  ]
End[]

EndPackage[]

